extern sim_stat_rec swapout_time;

#define DLEN    K(11)

volatile int serial_no=0;
volatile long long long_sum=0LL;
volatile long long short_sum=0LL;
struct timeval xbefore, xafter;
struct timeval ybefore, yafter;

char *hfmt= "#%7s %6s %12s %7s %8s %8s %7s\n";
char *fmt = "%8d %6d %12lld %7.1f %8d %8.2f %7.2f\n";

int report_interval=1000;
int nobj_thread=2000;


aLoop(int tno)
{
    int i, j;
    bucket_t *bk;
    skcache_entry_t *newone;
    skcache_entry_t *stored_pos;
    skcache_frag_t *cur;
    skcache_frag_t *last;
    int h1, h2, h3, h4;
    int already_have;
    int have, nothave;
    int chk;
    off_t olen;
    int cur_oid;
    int takes;
    
    char sample_data[2*DLEN];

    have = 0;
    nothave = 0;
    j = 0;
    newone = NULL;

    Trace("disk stresser# %d, booted\n", tno);

    for(i=0;i<nobj_thread;i++) {

        MUTEX_LOCK(&stat_lock);
        cur_oid = ++serial_no;
        MUTEX_UNLOCK(&stat_lock);

        h1 = rand() % radix_h1;
        h2 = rand() % radix_h2;
        h3 = rand() % radix_h3;
        h4 = rand() % radix_h4;

        bk = &skcache_bucket_base[h1];

        LOCK_BK(bk);

        newone = skcache_pull_new_entry(bk);
        if(newone==NULL) {
            Error("no memory");
            break;
        }

        newone->oid = cur_oid;
        newone->h1 = h1;
        newone->h2 = h2;
        newone->h3 = h3;
        newone->h4 = h4;
        sprintf(newone->name, "someURL%d", cur_oid);

        stored_pos =
            (skcache_entry_t*) sk_assign(&bk->cup, newone,
                &bk->elem_freelist, skcache_entry_cmp,
                &already_have, 0);

        if(stored_pos==NULL || already_have<0) {
            Trace("have %d,%d,%d,%d\n", h1, h2, h3, h4);

            skcache_push_idle_entry(bk, newone);

            have++;
            continue;
        }
        else {
            nothave++;
        }

        olen = DLEN + DLEN/2 - (rand()%DLEN);

        stored_pos->totallength= olen;
        stored_pos->m_recorded = NULL;
        stored_pos->d_recorded = 0;
        stored_pos->count = 0;          /* always new */
        stored_pos->ref   = 0;          /* always new */

        {
            int c;
            off_t tmp_len;
            off_t alen;

            c = 0;
            tmp_len = 0;
            last = NULL;

            alen = olen;
    
            do {
                cur = skcache_pull_new_frag(bk);
                if(!cur) {
                    goto loop_end;
                }

                if(last==NULL) {
                    stored_pos->m_recorded = cur;
                }
                else {
                    last->next = cur;
                }

                memcpy(cur->ent, &sample_data[tmp_len], MIN(cur->size, alen));
                cur->len = MIN(cur->size, alen);
                cur->next = NULL;

                alen -= MIN(cur->size, alen);

                c++;
                tmp_len += (off_t)cur->len;
                last = cur;
            } while(alen>0);

#if 0
            Trace("store oid# %d, %6ld bytes in %2d frags\n",
                cur_oid, (long)tmp_len, c);
#endif
        }

        if(stored_pos->m_recorded==NULL) {
            Trace("stored_pos %p, m_recorded %p\n",
                stored_pos, stored_pos->m_recorded);
        }
        
#if 1
        /* swapout and delete, normal process */
        chk = skcache_entry_swapout(bk, stored_pos);
        if(chk) {
            if(chk==XSK_DISKFULL) {
                Error("no space in filesystem, break");
                UNLOCK_BK(bk);
                break;
            }
            Error("fail swapout: reason %d", chk);
        }
        else {
            chk = skcache_entry_delete(bk, stored_pos);
        }
#else
        /* delete only */
        chk = skcache_entry_delete(bk, stored_pos);
#endif

        j++;

        UNLOCK_BK(bk);

        MUTEX_LOCK(&stat_lock);
        short_sum += (long long)olen;
        long_sum += (long long)olen;
        MUTEX_UNLOCK(&stat_lock);

    }

loop_end:

    Trace("disk stresser# %d, done ; have %d %.2f%%, not have %d\n",
        tno, 
        have, (have+nothave>0 ? have*100.0/(have+nothave) : -1),
        nothave);

    fflush(stderr);
    fflush(stdout);

    return;
}


void*
DiskJudge(void* dummy)
{
    int no;

    no = (int)dummy;
    aLoop(no);

    pthread_exit(NULL);
}

void*
LoadMonitor(void* dummy)
{
    int cur_oid;
    int showflag;
    int takes;
    int xi;
    int last_look;
    
    xi = 100;

    last_look = 0;
    while(1) {
        MUTEX_LOCK(&stat_lock);
        cur_oid  = serial_no;
        showflag = cur_oid>0 ?
            ((cur_oid-last_look)>report_interval ? 1 : 0 ) : 0;

        if(showflag) {
            gettimeofday(&yafter, NULL);
            takes = KOTETU_SUB_TIME(yafter, ybefore);
            fprintf(stdout, fmt, cur_oid, 1,
                (long long)long_sum, fdG(long_sum),
                takes, ((double)report_interval/takes)*1000.0,
                ((double) (short_sum)/takes/(1024*1024))*1000.0
                );

            gettimeofday(&ybefore, NULL);
            short_sum = 0LL;
            last_look = cur_oid;
        }

        MUTEX_UNLOCK(&stat_lock);
        fflush(stdout);

#if 0
        if(showflag) {
#if 0
            Trace("\tfrag use %d, free %d\n", frag_use, frag_free); 
#endif
            mon_fs(ybefore.tv_sec - boot_date);
        }
#endif

        if(xi>0) {
            msleep(xi);
        }
    }

    return;
}

void
DiskBench()
{
    int i;
    int chk;
    int takes;
    int n_bench_th=32;
    void *dish;
    pthread_t th_load;
    int nobj_total;
    bucket_t *bk;

    srand(time(NULL) & getpid());

    for(i=0;i<radix_h1;i++) {
        bk = &skcache_bucket_base[i];
        Trace("%3d:: bk %3d: %3d * %d + %3d <=> %3d\n",
            i, bk->bid, (bk->bid)/radix_disk,
            radix_disk, (bk->bid)%radix_disk, bk->fs);
    }


    fprintf(stdout, "\n"); 
    fprintf(stdout, hfmt, "serial", "lap",
        "size","", "take", "rate", "");
    fprintf(stdout, hfmt, "-", "-",
        "B","GB", "ms", "op/s", "MB/s");

    report_interval = 1000;
    n_bench_th  = 64;
    nobj_thread = 100000/n_bench_th;

    nobj_total = n_bench_th * nobj_thread;
    Trace("the total number of objects = %d (%d th x %d)\n",
        nobj_total, n_bench_th, nobj_thread);

    gettimeofday(&xbefore, NULL);
    gettimeofday(&ybefore, NULL);

    chk = pthread_create(&th_monitor, p_th_attr_mid,
            LoadMonitor, (void*)i);
    for(i=0;i<n_bench_th;i++) {
        chk = pthread_create(&th_load, p_th_attr_mid,
                DiskJudge, (void*)i);
    }

    while(1) {
        MUTEX_LOCK(&stat_lock);
        if(serial_no >= nobj_total) {
            MUTEX_UNLOCK(&stat_lock);
            break;
        }
        MUTEX_UNLOCK(&stat_lock);
        msleep(500);
    }

    gettimeofday(&yafter, NULL);
    gettimeofday(&xafter, NULL);

    msleep(500);

    {
        takes = KOTETU_SUB_TIME(yafter, ybefore);
        fprintf(stdout, fmt, serial_no, serial_no,
            (long long)long_sum, fdG(long_sum),
            takes, ((double)(serial_no%report_interval)/takes)*1000.0,
            ((double) (short_sum)/takes/(1024*1024))*1000.0
            );
    }

    takes = KOTETU_SUB_TIME(xafter, xbefore);

    fprintf(stdout, "\n");

    if(serial_no>0) {
        fprintf(stdout, "#");
        fprintf(stdout, fmt, serial_no, serial_no,
            (long long)long_sum, fdG(long_sum),
            takes, ((double)serial_no/takes)*1000.0);
        fprintf(stdout, "\n");
    }

    fflush(stderr);
    fflush(stdout);
#if 0
    sim_stat_show99_lock(&swapout_time, stdout);
#endif

    mon_fs(takes);
}


