#ifdef HAVE_CONFIG_H
#include "config.h"
#endif

#include <stdio.h>
#include <stdlib.h>
#include <pthread.h>

#include "mlist.h"

/*
#define NAMELEN     (32)
#define MLISTLEN    (16)
*/

typedef struct {
    char        name[MLIST_NAMELEN];
    void*       addr;
    long long   len;
} melem_t;

pthread_mutex_t mlist_lock;
int mlen=0;
melem_t mlist[MLIST_LISTLEN];


int
mlist_get_param(int *n, long long *len)
{
    int i, j;
    long long sum;

    pthread_mutex_lock(&mlist_lock);

    sum = 0LL;
    j = mlen;
    for(i=0;i<j;i++) {
        sum += mlist[i].len;
    }

    pthread_mutex_unlock(&mlist_lock);

    *n = j;
    *len = sum;

    return 0;
}

int
mlist_add_elem(char *xname, void* xaddr, long long xlen)
{
    int ret;

    pthread_mutex_lock(&mlist_lock);
    if(mlen<MLIST_LISTLEN) {
        strcpy(mlist[mlen].name, xname);
        mlist[mlen].addr = xaddr;
        mlist[mlen].len  = xlen;
        mlen++;
        ret = 0;
    }
    else {
        ret = 1;
    }
    pthread_mutex_unlock(&mlist_lock);

    return ret;
}

int
mlist_show()
{
    int i;
    long long sum, total;

    pthread_mutex_lock(&mlist_lock);
    
    fprintf(stdout, "memory blocks on heap area\n");

    sum = 0LL;
    for(i=0;i<mlen;i++) {
        sum += mlist[i].len;
    }
    total = sum;
        
    sum = 0LL;
    for(i=0;i<mlen;i++) {
        sum += mlist[i].len;
        
        fprintf(stdout, " %2d: %-22s 0x%08x %10lld %7.2fM %6.2f%% %10lld\n",
            i, mlist[i].name, mlist[i].addr, mlist[i].len,
            (double)mlist[i].len/(1024*1024),
            ((double)mlist[i].len*100.0/total),
            sum);
    }

    fprintf(stdout, "     %-22s            %10lld %7.2fM\n",
        "total", sum, (double)sum/(1024*1024));
    fflush(stdout);

    pthread_mutex_unlock(&mlist_lock);
}

long long
mlist_get_total()
{
    int i;
    long long sum;

    pthread_mutex_lock(&mlist_lock);
    sum = 0LL;
    for(i=0;i<mlen;i++) {
        sum += mlist[i].len;
    }
    pthread_mutex_unlock(&mlist_lock);

    return sum;
}

