/*
 * code-name 'KOTETU'
 *
 * sess.h: definition for sessions
 *
 *      by k-chinen@is.aist-nara.ac.jp, 1999
 *
 * $Id$
 */

#ifndef _SESS_H_
#define _SESS_H_


#ifndef NSESS_TOTAL_MAX
#define NSESS_TOTAL_MAX          (128000)
#endif

#ifndef N_HANDLER_MAX
#define N_HANDLER_MAX       (128)
#endif
#ifndef N_SWAPPER_MAX
#define N_SWAPPER_MAX       N_HANDLER_MAX
#endif

#define STATE_IDLE              (0)
#define STATE_ACCEPT            (1)
#define STATE_RECV_REQ          (2)
#define STATE_THINK             (3)
#define STATE_REPORT            (4)
#define STATE_WAIT_FRIEND       (5)
#define STATE_SEND_CACHE        (6)
#define STATE_QUERY_ADDR        (7)
#define STATE_WAIT_ADDR         (8)
#define STATE_CONN_SERV         (9)
#define STATE_WAIT_CONN_SERV    (10)
#define STATE_SEND_REQUEST      (11)
#define STATE_RECV_RES          (12)
#define STATE_BODYPARSE         (13)
#define STATE_CLOSE             (14)
#define STATE_DONE              (15)
#define STATE_MESSAGE           (16)
#define STATE_RELAY_DIRECT      (17)
#define STATE_WAIT_SWAPIN       (18)
#define STATE_WAIT_SWAPOUT      (19)
#define STATE_WAIT_SWAPLOAD     (20)
#define STATE_WAIT_DISK_GC      (21)
#define STATE_WAIT_MEMORY_GC    (22)
#define STATE_DUMMY_END         (23)

#define SESS_MISS               (0)
#define SESS_HIT                (1)

#define SESS_UNCACHABLE         (1)
#define SESS_CACHABLE           (2)

#define SESS_RELOAD             (4)
#define SESS_NORELOAD           (8)

#define SESS_PRIVATE            (16)
#define SESS_PUBLIC             (32)

#define MODE_CLIENT             (1)
#define MODE_SELF               (2)

#ifndef SESSION_TOTAL_TIMEOUT
#define SESSION_TOTAL_TIMEOUT   (300)
#endif
#ifndef RECEPT_TIMEOUT
#define RECEPT_TIMEOUT          (10)
#endif
#ifndef SEND_TIMEOUT
#define SEND_TIMEOUT            (30)
#endif
#ifndef FETCH_TIMEOUT
#define FETCH_TIMEOUT           (90)
#endif
#ifndef WFRIEND_TIMEOUT
#define WFRIEND_TIMEOUT         (60)
#endif
#ifndef ELSE_TIMEOUT
#define ELSE_TIMEOUT            (60)
#endif

#ifndef ABS_PATH    
#define ABS_PATH        (1)
#define REL_PATH        (2)
#endif

#ifndef NSESS_HANDLER_MAX
#define NSESS_HANDLER_MAX (512)
#endif /* NSESS_HANDLER_MAX */


typedef struct _session {
    pthread_mutex_t busy;
    struct _session *prev, *next;

    int  sid;
    int  psid;
    int  mode;
    int  level;
    int  state;
    int  reason;
    int  reason2;

    int  tic;
    int  hf;                        /* history of friend */

    struct timeval accept_time;

    time_t last_action;
    int  reported;

    int  method;
    int  rcode;                     /* response code */
    int  count;

    int  cfd;                       /* http client */
    char clientname[L_HOST];
    int  cver_major, cver_minor;

    int  rfd;                       /* hostname resolver */
    int  rcount;
    struct timeval r_start;

    int  sfd;                       /* http server */
    char servername[L_HOST];
    int  serverport;
    char serverpath[L_URL];
    int  sver_major, sver_minor;

    struct sockaddr_in serveraddr;

    char reqbuf[L_REQBUF];
    off_t reqlen;
    time_t ims;

    int cachable;
    int reload;
    int ishit;
    int publish;

    int  fhash;
    int  shash;
    int  thash;
    int  ffhash;
    char loc[L_URL];
    int  abs_or_rel;    /*
                         * absolute or relative path in URL.
                         *      if received path is relative path,
                         *      this program guess "I am reverse proxy".
                         *
                         *      unknown     0
                         *      absolute    1
                         *      relative    2
                         */

    void *bucket;
    void *cache_entry;
/*
    bucket_t *bk;
    skcache_entry_t *cache_entry;
*/

    int hlen;

    int  oid;
    char *buf;
    off_t pos;
    off_t len;
    time_t lmd;
    time_t expire;
    off_t clen;

    off_t cread, cwrite;        /* client side read/write */
    off_t sread, swrite;        /* server side read/write */

    void *follow;
} sess_t;

typedef struct {
    int sid;
    char loc[L_URL];
} prompt_t;


extern pthread_mutex_t sess_lock;
extern size_t sess_size;
extern int nhot;

extern sim_stat_rec rcode_stat;

extern sim_stat_rec ret_time;
extern sim_stat_rec ret_time_short;

extern sim_stat_rec len_perread_stat;
extern sim_stat_rec object_size_stat;
extern sim_stat_rec url_first_hash_stat;
extern sim_stat_rec url_second_hash_stat;
extern sim_stat_rec url_third_hash_stat;
extern sim_stat_rec url_fourth_hash_stat;


extern pthread_mutex_t stat_lock;

extern time_t boot_date;
extern time_t service_start_date;

extern volatile long long request_sirial_count;
extern volatile int receive_request_count;
extern volatile int process_request_count;
extern volatile int issue_prefetch_count;
extern volatile int report_request_count;
extern volatile int reject_request_count;
extern volatile int ignore_request_count;

extern volatile int send_reply_count;
extern volatile int hit_reply_count;
extern volatile int miss_reply_count;
extern volatile int error_reply_count;

extern volatile int hash_colision;
extern volatile int compaction_count;
extern volatile int expire_count;

extern int read_reply_total;
extern int read_reply_success;
extern int read_reply_fail;

extern int cache_newover_count;
extern int cache_newfree_count;
extern int cache_newmin_count;


extern volatile long long intra_incoming;
extern volatile long long extra_outgoing;
extern volatile long long extra_incoming;
extern volatile long long intra_outgoing;


extern char *MSG_STATE_[];





#endif /* _SESS_H_ */
