/*
 * simple stat routines
 *          by k-chinen@is.aist-nara.ac.jp, 1999
 *
 * $Id$
 */

#ifdef HAVE_CONFIG_H
#include "config.h"
#endif

#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <pthread.h>

#include <errno.h>

#include "sim_stat.h"
#include "mem.h"
#include "msg.h"

int
sim_stat_clear(sim_stat_rec *srec)
{
    int i;
    int chk;

    chk = MUTEX_LOCK(&srec->busy);

    for(i=0;i<srec->nslot;i++) {
        srec->hcount[i] = 0L;
    }
    srec->count = 0L;
    srec->igcount = 0L;
    srec->vsum = 0L;

    chk = MUTEX_UNLOCK(&srec->busy);

    return 0;
}

int
sim_stat_U_alloc_new(sim_stat_rec *addr, char *name, int un, int sz, long *ar)
{
    sim_stat_rec *ret;
    int chk;

    if(addr==NULL) {
        return -1;
    }
    ret = addr;

    chk = pthread_mutex_init(&addr->busy, NULL);

    chk = MUTEX_LOCK(&addr->busy);
    strcpy(ret->name, name);
    ret->vunit = un;
    ret->nslot = sz;
    ret->hcount = ar;

    chk = MUTEX_UNLOCK(&addr->busy);

    sim_stat_clear(ret);

    return 0;
}

sim_stat_rec*
sim_stat_U_new(char *name, int un, int sz)
{
    int chk;
    sim_stat_rec *ret;

    ret = (sim_stat_rec*) malloc(sizeof(sim_stat_rec));
    if(ret==NULL) {
        return NULL;
    }
    MLOCK((void*)ret ,sizeof(sim_stat_rec));

    chk = pthread_mutex_init(&ret->busy, NULL);

    chk = MUTEX_LOCK(&ret->busy);
    strcpy(ret->name, name);
    ret->vunit = un;
    ret->nslot = sz;
    ret->hcount = (long*) malloc(sizeof(long)*sz);
    if(ret->hcount==NULL) {
        ret = NULL;
    }
    MLOCK((void*)ret->hcount, sizeof(long)*sz);

    chk = MUTEX_UNLOCK(&ret->busy);

    sim_stat_clear(ret);

    return ret;
}

int
sim_stat_alloc_new(sim_stat_rec *addr, char *name, int sz, long *ar)
{
    return sim_stat_U_alloc_new(addr, name, 1, sz, ar);
}

sim_stat_rec*
sim_stat_new(char *name, int sz)
{
    return sim_stat_U_new(name, 1, sz);
}



int
sim_stat_show(sim_stat_rec *rec, FILE *fp)
{
    int i;
    int chk;
    long tsum, asum;

    chk = MUTEX_LOCK(&rec->busy);

    fprintf(fp, "=== %s ===\n", rec->name);

    if(rec->count<=0) {
        fprintf(fp, "no sample\n");
        chk = MUTEX_UNLOCK(&rec->busy);
        return -1;
    }


    fprintf(fp, "# access %d time, sum %d -> ave %.3f\n",
        rec->count, rec->vsum,
        rec->count<=0 ? -1.0 : (double)rec->vsum/rec->count);

    tsum = 0L;
    for(i=0;i<rec->nslot;i++) {
        tsum += rec->hcount[i];
    }
    tsum += rec->igcount;
    asum = tsum;

    fprintf(fp, "#%6s %9s %6s %9s %6s\n",
            "value", "count", "PDF%", "sum", "CDF%");
    tsum = 0L;
    for(i=0;i<rec->nslot;i++) {
        tsum += rec->hcount[i];
        if(rec->hcount[i]>0) {
            fprintf(fp, " %6d %9d %6.2f %9d %6.2f\n",
                i, rec->hcount[i],
                asum<=0 ? -1.0 : 100.0*(double)rec->hcount[i]/asum,
                tsum, 
                asum<=0 ? -1.0 : 100.0*(double)tsum/asum);
        }
    }

    tsum += rec->igcount;
    fprintf(fp, " %6s %9d %6.2f %9d %6.2f\n",
        "others", rec->igcount,
        asum<=0 ? -1.0 : 100.0*(double)rec->igcount/asum,
        tsum, 
        asum<=0 ? -1.0 : 100.0*(double)tsum/asum);


    fprintf(fp, "\n");

    chk = MUTEX_UNLOCK(&rec->busy);

    return 0;
}

int
sim_stat_show_lock(sim_stat_rec *rec, FILE *fp)
{
    int i;
    int chk;
    long tsum, asum;


    fprintf(fp, "=== %s ===\n", rec->name);

    if(rec->count<=0) {
        fprintf(fp, "no sample\n");
        return -1;
    }


    fprintf(fp, "# access %d time, sum %d -> ave %.3f\n",
        rec->count, rec->vsum, (double)rec->vsum/rec->count);

    tsum = 0L;
    for(i=0;i<rec->nslot;i++) {
        tsum += rec->hcount[i];
    }
    tsum += rec->igcount;
    asum = tsum;

    fprintf(fp, "#%6s %9s %6s %9s %6s\n",
            "value", "count", "PDF%", "sum", "CDF%");
    tsum = 0L;
    for(i=0;i<rec->nslot;i++) {
        tsum += rec->hcount[i];
        if(rec->hcount[i]>0) {
            fprintf(fp, " %6d %9d %6.2f %9d %6.2f\n",
                i, rec->hcount[i],
                asum<=0 ? -1.0 : 100.0*(double)rec->hcount[i]/asum,
                tsum,
                asum<=0 ? -1.0 : 100.0*(double)tsum/asum);
        }
    }

    tsum += rec->igcount;
    fprintf(fp, " %6s %9d %6.2f %9d %6.2f\n",
        "others", rec->igcount,
        asum<=0 ? -1.0 : 100.0*(double)rec->igcount/asum,
        tsum,
        asum<=0 ? -1.0 : 100.0*(double)tsum/asum);


    fprintf(fp, "\n");

    return 0;
}

int
sim_stat_showX_lock(sim_stat_rec *rec, FILE *fp)
{
    int i;
    int chk;
    long tsum, asum;


    fprintf(fp, "=== %s ===\n", rec->name);

    if(rec->count<=0) {
        fprintf(fp, "no sample\n");
        return -1;
    }


    fprintf(fp, "# access %d time, sum %d -> ave %.3f\n",
        rec->count, rec->vsum,
        rec->count<=0 ? -1.0 : (double)rec->vsum/rec->count);

    tsum = 0L;
    for(i=0;i<rec->nslot;i++) {
        tsum += rec->hcount[i];
    }
    tsum += rec->igcount;
    asum = tsum;
    if(asum<=0) {
        fprintf(fp, "# target is empty or error\n");
        goto end;
    }

    fprintf(fp, "#%8s %9s %6s %9s %6s\n",
            "value", "count", "PDF%", "sum", "CDF%");
    tsum = 0L;
    for(i=0;i<rec->nslot;i++) {
        tsum += rec->hcount[i];
        if(rec->hcount[i]>0) {
            fprintf(fp, " %8x %9d %6.2f %9d %6.2f\n",
                i, rec->hcount[i], 100.0*(double)rec->hcount[i]/asum,
                tsum, 100.0*(double)tsum/asum);
        }
    }

    tsum += rec->igcount;
    fprintf(fp, " %8s %9d %6.2f %9d %6.2f\n",
        "others", rec->igcount, 100.0*(double)rec->igcount/asum,
         tsum, 100.0*(double)tsum/asum);


end:
    fprintf(fp, "\n");

    return 0;
}

int
sim_stat_show_R_lock(sim_stat_rec *rec, double R, FILE *fp)
{
    int i;
    int chk;
    long tsum, asum;

    if(rec->count<=0) {
        fprintf(fp, "no sample in '%s'\n", rec->name);
        return -1;
    }
    if(rec->vunit==1) {
        fprintf(fp, "=== %s === (until CDF %.2f%%)\n", rec->name, R);
    }
    else {
        fprintf(fp, "=== %s === <unit %d> (until CDF %.2f%%)\n",
            rec->name, rec->vunit, R);
    }

    fprintf(fp, "# access %d time, sum %d -> ave %.3f\n",
        rec->count, rec->vsum*rec->vunit,
        rec->count<=0 ? -1.0 : ((double)rec->vsum*rec->vunit)/rec->count);

    tsum = 0L;
    for(i=0;i<rec->nslot;i++) {
        tsum += rec->hcount[i];
    }
    tsum += rec->igcount;
    asum = tsum;
    if(asum<=0) {
        fprintf(fp, "# target is empty or error\n");
        goto end;
    }

    fprintf(fp, "#%6s %9s %6s %9s %6s\n",
            "value", "count", "PDF%", "sum", "CDF%");
    tsum = 0L;
    for(i=0;i<rec->nslot;i++) {
        tsum += rec->hcount[i];
        if(rec->hcount[i]>0) {
            fprintf(fp, " %6d %9d %6.2f %9d %6.2f\n",
                i*rec->vunit,
                rec->hcount[i], 100.0*(double)rec->hcount[i]/asum,
                tsum, 100.0*(double)tsum/asum);
        }
        if(100.0*(double)tsum/asum>=R) {
            break;
        }
    }

end:
    fprintf(fp, "=== END\n\n");

    return 0;
}

int
sim_stat_show995_lock(sim_stat_rec *rec, FILE *fp)
{
    return sim_stat_show_R_lock(rec, 99.50, fp);
}
int
sim_stat_show99_lock(sim_stat_rec *rec, FILE *fp)
{
    return sim_stat_show_R_lock(rec, 99.00, fp);
}
int
sim_stat_show95_lock(sim_stat_rec *rec, FILE *fp)
{
    return sim_stat_show_R_lock(rec, 95.00, fp);
}
int
sim_stat_show91_lock(sim_stat_rec *rec, FILE *fp)
{
    return sim_stat_show_R_lock(rec, 91.00, fp);
}
int
sim_stat_show90_lock(sim_stat_rec *rec, FILE *fp)
{
    return sim_stat_show_R_lock(rec, 90.00, fp);
}

int
sim_stat_show_lock_fd(sim_stat_rec *rec, int ofd)
{
    int i;
    int chk;
    long tsum, asum;


    fdprintf(ofd, "=== %s === (without lock)\n", rec->name);

    if(rec->count<=0) {
        fdprintf(ofd, "no sample\n");
        return -1;
    }


    fdprintf(ofd, "# access %d time, sum %d -> ave %.3f\n",
        rec->count, rec->vsum, (double)rec->vsum/rec->count);

    tsum = 0L;
    for(i=0;i<rec->nslot;i++) {
        tsum += rec->hcount[i];
    }
    tsum += rec->igcount;
    asum = tsum;

    fdprintf(ofd, "#%6s %9s %6s %9s %6s\n",
            "value", "count", "PDF%", "sum", "CDF%");
    tsum = 0L;
    for(i=0;i<rec->nslot;i++) {
        tsum += rec->hcount[i];
        if(rec->hcount[i]>0) {
            fdprintf(ofd, " %6d %9d %6.2f %9d %6.2f\n",
                i, rec->hcount[i], 100.0*(double)rec->hcount[i]/asum,
                tsum, 100.0*(double)tsum/asum);
        }
    }

    tsum += rec->igcount;
    fdprintf(ofd, "#%6s %9d %6.2f %9d %6.2f\n",
        "others", rec->igcount, 100.0*(double)rec->igcount/asum,
         tsum, 100.0*(double)tsum/asum);


    fdprintf(ofd, "\n");

    return 0;
}

int
sim_stat_show_HTML_nolock_fd(sim_stat_rec *rec, int ofd)
{
    int i;
    int chk;
    long tsum, asum;
    int v;


    fdprintf(ofd, "<table border=\"0\" cellspacing=\"0\" cellpadding=\"2\">\n\
<tr bgcolor=\"#ff9999\"> <th colspan=\"7\">%s (without lock)\n", rec->name);


    if(rec->count<=0) {
        fdprintf(ofd, "<tr><th>no sample\n");
        goto out;
    }


    tsum = 0L;
    for(i=0;i<rec->nslot;i++) {
        tsum += rec->hcount[i];
    }
    tsum += rec->igcount;
    asum = tsum;

    fdprintf(ofd, "<tr bgcolor=\"#ffcccc\">\n\
<th>%s <th>%s <th>%s <th>%s <th>%s <th width=\"390\">%s <th width=\"390\">%s\n",
            "value", "count", "PDF%", "sum", "CDF%",
            "graph of PDF", "graph of CDF");
    tsum = 0L;
    for(i=0;i<rec->nslot;i++) {
        tsum += rec->hcount[i];
        if(rec->hcount[i]>0) {
            fdprintf(ofd,
"<tr><td align=\"right\"> %d \
<td align=\"right\"> %d \
<td align=\"right\"> %6.2f \
<td align=\"right\"> %d \
<td align=\"right\"> %6.2f  ",
                i, rec->hcount[i],
                asum<=0 ? -1.0 : 100.0*(double)rec->hcount[i]/asum,
                tsum, asum<=0 ? -1.0 : 100.0*(double)tsum/asum);

            if(asum<=0)
                v = -1.0;
            else
                v = (int)(100.0*(double)rec->hcount[i]/asum);
            if(v==100) {
                fdprintf(ofd,
"<td width=\"390\"><img width=\"200\" src=\"/img/bw200p100.jpg\"> 100  ");
            }
            else
            if(v<=0) {
                fdprintf(ofd,
"<td width=\"390\"> 0  ");
            }
            else
            if(v<=10 || v%10==0) {
                fdprintf(ofd,
"<td width=\"390\"><img width=\"%d\" src=\"/img/bw200p%02d.jpg\"> %d  ",
                    v*2, v, v);
            }
            else {
                fdprintf(ofd,
"<td width=\"390\">\
<img width=\"%d\" src=\"/img/bw200p%02d.jpg\">\
<img width=\"%d\" src=\"/img/bw200p%02d.jpg\"> %d  ",
                    (v/10)*20, (v/10)*10, (v%10)*2, v%10 , v);
            }


            if(asum<=0)
                v = -1.0;
            else
                v = (int)(100.0*(double)tsum/asum);
            if(v==100) {
                fdprintf(ofd,
"<td width=\"390\"><img width=\"200\" src=\"/img/bw200p100.jpg\"> 100  ");
            }
            else
            if(v<=0) {
                fdprintf(ofd,
"<td width=\"390\"> 0  ");
            }
            else
            if(v<=10 || v%10==0) {
                fdprintf(ofd,
"<td width=\"390\"><img width=\"%d\" src=\"/img/bw200p%02d.jpg\"> %d  ",
                    v*2, v, v);
            }
            else {
                fdprintf(ofd,
"<td width=\"390\"><img width=\"%d\" src=\"/img/bw200p%02d.jpg\">\
<img width=\"%d\" src=\"/img/bw200p%02d.jpg\"> %d  ",
                    (v/10)*20, (v/10)*10, (v%10)*2, v%10 , v);
            }

            fdprintf(ofd, "\n");
        }
    }

    if(rec->igcount>0) {
        tsum += rec->igcount;
        fdprintf(ofd,
"<tr><td align=\"center\"> %s\n\
<td align=\"right\"> %d\n\
<td align=\"right\"> %6.2f\n\
<td align=\"right\"> %d\n\
<td align=\"right\"> %6.2f\n",
            "others", rec->igcount, 
            asum<=0 ? -1.0 : 100.0*(double)rec->igcount/asum,
            tsum, asum<=0 ? -1.0 : 100.0*(double)tsum/asum);

        if(asum<=0)
            v = -1.0;
        else
            v = (int)(100.0*(double)rec->igcount/asum);
        if(v==100) {
            fdprintf(ofd, "<td><img width=\"200\" src=\"/img/bw200p100.jpg\"> 100  ");
        }
        else
        if(v<=0) {
            fdprintf(ofd, "<td> 0  ");
        }
        else
        if(v<=10 || v%10==0) {
            fdprintf(ofd, "<td><img width=\"%d\" src=\"/img/bw200p%02d.jpg\"> %d  ",
                v*2, v, v);
        }
        else {
            fdprintf(ofd,
"<td><img width=\"%d\" src=\"/img/bw200p%02d.jpg\"><img width=\"%d\" src=\"/img/bw200p%02d.jpg\"> %d  ",
                (v/10)*20, (v/10)*10, (v%10)*2, v%10, v);
        }

        if(asum<=0)
            v = -1.0;
        else
            v = (int)(100.0*(double)tsum/asum);
        if(v==100) {
            fdprintf(ofd, "<td><img width=\"200\" src=\"/img/bw200p100.jpg\"> 100  ");
        }
        else
        if(v<=0) {
            fdprintf(ofd, "<td> 0  ");
        }
        else
        if(v<=10 || v%10==0) {
            fdprintf(ofd, "<td><img width=\"%d\" src=\"/img/bw200p%02d.jpg\"> %d  ",
                v*2, v, v);
        }
        else {
            fdprintf(ofd,
"<td><img width=\"%d\" src=\"/img/bw200p%02d.jpg\"><img width=\"%d\" src=\"/img/bw200p%02d.jpg\"> %d  ",
                (v/10)*20, (v/10)*10, (v%10)*2, v%10, v);
        }

        fdprintf(ofd, "\n");

    }


    fdprintf(ofd,
"<tr><th bgcolor=\"#ffcccc\">total <td align=\"right\"> %d\n\
<tr><th bgcolor=\"#ffcccc\">average <td align=\"right\"> %d\n\
<tr><th bgcolor=\"#ffcccc\">average <td align=\"right\"> %.1f\n",
        rec->count, rec->vsum, (double)rec->vsum/rec->count);

out:
    fdprintf(ofd, "</table>\n");

    return 0;
}

int
sim_stat_show_skip_lock_fd(sim_stat_rec *rec, int ofd, int sk)
{
    int i;
    int chk;
    long tsum, asum;


    fdprintf(ofd, "=== %s === (with skip without lock)\n", rec->name);

    if(rec->count<=0) {
        fdprintf(ofd, "no sample\n");
        return -1;
    }


    fdprintf(ofd, "# access %d time, sum %d -> ave %.3f\n",
        rec->count, rec->vsum,
        rec->count<=0 ? -1.0 : (double)rec->vsum/rec->count);

    tsum = 0L;
    for(i=0;i<rec->nslot;i++) {
        tsum += rec->hcount[i];
    }
    tsum += rec->igcount;
    asum = tsum;

    fdprintf(ofd, "#%6s %9s %6s %9s %6s\n",
            "value", "count", "PDF%", "sum", "CDF%");
    tsum = 0L;
    for(i=0;i<rec->nslot;i++) {
        tsum += rec->hcount[i];
#if 0
        if(rec->hcount[i]>0) {
            fdprintf(ofd, " %6d %9d %6.2f %9d %6.2f\n",
                i, rec->hcount[i], 100.0*(double)rec->hcount[i]/asum,
                tsum, 100.0*(double)tsum/asum);
        }
#else
        if(i%sk==0 && tsum>0) {
            fdprintf(ofd, " %6d %9d %6.2f %9d %6.2f\n",
                i, rec->hcount[i],
                asum<=0 ? -1.0 : 100.0*(double)rec->hcount[i]/asum,
                tsum,
                asum<=0 ? -1.0 : 100.0*(double)tsum/asum);
        }
#endif
    }

    tsum += rec->igcount;
    fdprintf(ofd, "#%6s %9d %6.2f %9d %6.2f\n",
        "others", rec->igcount, 
        asum<=0 ? -1.0 : 100.0*(double)rec->igcount/asum,
        tsum,
        asum<=0 ? -1.0 : 100.0*(double)tsum/asum);


    fdprintf(ofd, "\n");

    return 0;
}

int
sim_stat_show_short(sim_stat_rec *rec, FILE *fp)
{
    int chk;

    chk = MUTEX_LOCK(&rec->busy);
    fprintf(fp, "access %d time, sum %d -> ave %.3f\n",
        rec->count, rec->vsum, 
        rec->count<=0 ? -1.0 : (double)rec->vsum/rec->count);
    chk = MUTEX_UNLOCK(&rec->busy);

    return 0;
}

int
sim_stat_reset(sim_stat_rec *rec)
{
    int i;
    int chk;

    chk = MUTEX_LOCK(&rec->busy);

    rec->vsum = 0L;
    rec->count = 0L;
    for(i=0;i<rec->nslot;i++) {
        rec->hcount[i] = 0L;
    }
    rec->igcount = 0L;
    chk = MUTEX_UNLOCK(&rec->busy);

    return 0;
}

int
sim_stat_add_value(sim_stat_rec *rec, long cv)
{
    int chk;
    int xv;

    xv = cv/rec->vunit;

    chk = MUTEX_LOCK(&rec->busy);

    rec->count++;
    rec->vsum += xv;
    if(xv>=0 && xv< rec->nslot) {
        rec->hcount[xv]++;
    }
    else {
        rec->igcount++;
    }
    chk = MUTEX_UNLOCK(&rec->busy);

    return 0;
}

double
sim_stat_get_ave(sim_stat_rec *rec)
{
    int chk;
    double ret;

    chk = MUTEX_LOCK(&rec->busy);
    if(rec->count<=0) {
        ret = -1.0;
    }
    else {
        ret = (double)rec->vsum*rec->vunit/rec->count;
    }
    chk = MUTEX_UNLOCK(&rec->busy);

    return ret;
}

